# frozen_string_literal: true

module Heptapod::NativeMercurial
  # Native Mercurial Migrator
  #
  # This is responsible for migrating a mercurial project from hg_git backend to native hg.
  class Migrator
    SHA_MAX_LEN = 40
    SHA_REGEX = '\b[0-9a-f]{7,40}\b'

    def initialize(project)
      @project = project
    end

    def vcs_type_enforcer
      proj_initial_vcs = @project.vcs_type
      if @project.vcs_type == 'hg'
        @project.vcs_type = 'hg_git'
        @project.save!
      end

      yield
    ensure
      # ensure that native project remains native in case of errors
      if proj_initial_vcs == 'hg'
        @project.vcs_type = 'hg'
        @project.save!
      end
    end

    def init_sha_map
      @hg_sha_map = @project.repository.hg_sha_map
    end

    def migrate!
      vcs_type_enforcer do
        init_sha_map
        do_migrate
      end
    end

    def log_error(err, msg)
      Gitlab::AppLogger.error(
        message: msg,
        backtrace: Gitlab::BacktraceCleaner.clean_backtrace(err.backtrace),
        project_id: @project.id,
        project_full_path: @project.full_path
      )
    end

    def do_migrate
      Gitlab::AppLogger.info(
        message: "Starting native hg migration",
        project_id: @project.id,
        project_full_path: @project.full_path
      )

      migrate_issues
      migrate_merge_requests
      migrate_pipelines
      migrate_notes
      migrate_releases

      # ensure all gitlab specific state files are present
      output, status = @project.repository.hg_call(
        ["hpd-ensure-all-gitlab-specific-state-files"], force_system_user: true
      )
      raise StandardError, output.strip unless status == 0

      @project.vcs_type = 'hg'
      @project.save!
    rescue StandardError => err
      msg = "Native mercurial migration failed: #{err}"
      log_error(err, msg)
      raise
    end

    def migrate_issues
      @project.issues.each do |issue|
        issue.title = replace_git_shas_with_hg(issue.title)
        issue.title_html = nil
        issue.description = replace_git_shas_with_hg(issue.description)
        issue.description_html = nil
        issue.save!
      end
    end

    def migrate_pipelines
      @project.all_pipelines.each do |pline|
        pline.sha = replace_git_shas_with_hg(pline.sha)
        pline.before_sha = replace_git_shas_with_hg(pline.before_sha)
        pline.source_sha = replace_git_shas_with_hg(pline.source_sha)
        pline.target_sha = replace_git_shas_with_hg(pline.target_sha)
        pline.save!
      end
    end

    # Note fields whose change should trigger a database save
    def note_watched_fields(note)
      [
         note.note, note.commit_id,
         note.original_position, note.position, note.change_position
      ]
    end

    def save_note_if_changed(note)
      before = note_watched_fields(note)
      yield
      after = note_watched_fields(note)

      return if after == before

      note.note_html = nil
      note.importing = true
      note.hg_native_migrating = true
      note.save!
    end

    def migrate_notes
      @project.notes.each do |note|
        save_note_if_changed(note) do
          note.note = replace_git_shas_with_hg(note.note)
          migrate_note_positions(note)
          note.commit_id = replace_git_shas_with_hg(note.commit_id)
        end
      rescue StandardError => err
        log_error(err, "Failed to migrate Note id=#{note.id}")
        raise
      end
    end

    def migrate_note_positions(note)
      if note.original_position.instance_of? Gitlab::Diff::Position
        note.original_position = _migrated_position(note.original_position)
      end

      if note.position.instance_of? Gitlab::Diff::Position
        note.position = _migrated_position(note.position)
      end

      if note.change_position.instance_of? Gitlab::Diff::Position
        note.change_position = _migrated_position(note.change_position)
      end
    end

    def _migrated_position(pos)
      pos_attrs = pos.to_h
      pos_attrs[:base_sha] = replace_git_shas_with_hg(pos_attrs[:base_sha])
      pos_attrs[:start_sha] = replace_git_shas_with_hg(
        pos_attrs[:start_sha])
      pos_attrs[:head_sha] = replace_git_shas_with_hg(pos_attrs[:head_sha])
      Gitlab::Diff::Position.new(pos_attrs)
    end

    def migrate_releases
      @project.releases.each do |rl|
        rl.sha = replace_git_shas_with_hg(rl.sha)
        rl.description = replace_git_shas_with_hg(rl.description)
        rl.description_html = nil
        rl.save!
      end
    end

    def migrate_merge_requests
      @project.merge_requests.each do |mr|
        mr.title = replace_git_shas_with_hg(mr.title)
        mr.title_html = nil
        mr.description = replace_git_shas_with_hg(mr.description)
        mr.description_html = nil

        # update all the columns ending with '_sha'
        mr.merge_commit_sha = replace_git_shas_with_hg(mr.merge_commit_sha)
        mr.rebase_commit_sha = replace_git_shas_with_hg(mr.rebase_commit_sha)
        mr.in_progress_merge_commit_sha = (
          replace_git_shas_with_hg(mr.in_progress_merge_commit_sha)
        )
        mr.squash_commit_sha = replace_git_shas_with_hg(mr.squash_commit_sha)
        mr.merge_ref_sha = replace_git_shas_with_hg(mr.merge_ref_sha)
        migrate_merge_request_diffs(mr)
        mr.save!
      end
    end

    # rubocop: disable CodeReuse/ActiveRecord
    def migrate_merge_request_diffs(mr)
      mr.merge_request_diffs.each do |diff|
        diff.base_commit_sha = replace_git_shas_with_hg(diff.base_commit_sha)
        diff.head_commit_sha = replace_git_shas_with_hg(diff.head_commit_sha)
        diff.start_commit_sha = replace_git_shas_with_hg(diff.start_commit_sha)
        diff.save!
      end

      # The table for MergeRequestDiffCommit has a composite primary key:
      # `PRIMARY KEY, btree (merge_request_diff_id, relative_order)`,
      # preventing updates using the normal ActiveRecord API, that only
      # supports single column primary keys.
      # Hence we have to go low level and build a query by hand with Arel,
      # the SQL query builder and AST representation used by ActiveRecord.
      # The crucial fact here is that the WHERE clauses pin down the two
      # columns that are part of the primary key (as of this writing).
      arel_table = MergeRequestDiffCommit.arel_table
      mr.all_commits.each do |mr_diff_commit|
        hg_sha = replace_git_shas_with_hg(mr_diff_commit.sha)
        hg_sha = "\\x#{hg_sha}"
        query = Arel::UpdateManager.new
          .table(arel_table)
          .set([[arel_table[:sha], hg_sha]])
          .where(arel_table[:merge_request_diff_id].eq(mr_diff_commit.merge_request_diff_id))
          .where(arel_table[:relative_order].eq(mr_diff_commit.relative_order))
          .to_sql
        MergeRequestDiffCommit.connection.execute(query)
      end
    end
    # rubocop: enable CodeReuse/ActiveRecord

    private

    def replace_git_shas_with_hg(text)
      return unless text.present? && @hg_sha_map.present?

      git_shas = @hg_sha_map.keys
      text.gsub(/#{SHA_REGEX}/) do |match|
        if match.length == SHA_MAX_LEN && git_shas.include?(match)
          @hg_sha_map[match]
        else
          full_git_sha = lambda {
            git_shas.each do |sha|
              if sha.starts_with? match
                return sha
              end
            end
            nil
          }.call
          full_git_sha.nil? ? match : @hg_sha_map[full_git_sha]
        end
      end
    end
  end
end
