# frozen_string_literal: true

require 'spec_helper'

RSpec.describe Heptapod::NativeMercurial::Migrator do
  let(:project) { create(:hg_git_project, :hg_repository)}
  let(:repo) { project.repository }
  let(:user) { build(:user) }
  let(:prepare_repo) do
    repo.hg_call(%w[up 0])
    hg_commit(1)
    repo.hg_call(%w[up 0])
    hg_commit(2, topic: 'feature')
  end

  let(:hg_sha) { repo.hgsha_from_rev('feature') }
  let(:git_sha) { repo.sha_from_hgsha(hg_sha) }
  let(:hg_base_sha) { repo.hgsha_from_rev('default') }
  let(:git_base_sha) { repo.sha_from_hgsha(hg_base_sha) }

  subject do
    sub = described_class.new(project)
    sub.init_sha_map
    sub
  end

  def hg_commit(cid, topic: nil)
    if topic
      repo.hg_call(['topic', topic])
    end

    repo.write_file(repo.hg_full_path + "/file#{cid}", "line\n")
    repo.hg_call(["add", "file#{cid}"])
    repo.hg_call(['ci', '-m', "Commit #{cid}", '--config',
                  'experimental.single-head-per-branch=False'],
                 user: user, for_write: true)
    repo.hg_git_invalidate_maps!
  end

  before do
    prepare_repo
  end

  context 'Issues' do
    let!(:issue) do
      create(:issue,
        project: project,
        description: "Sample description with Git commit: #{git_sha}",
        title: "Sample title for Git commit: #{git_sha}")
    end

    it 'migrate title and description' do
      subject.migrate_issues
      issue.reload
      expect(issue.description).to eq("Sample description with Git commit: #{hg_sha}")
      expect(issue.title).to eq("Sample title for Git commit: #{hg_sha}")
    end

    context 'Notes' do
      let!(:my_comment) do
        create(:note_on_issue, project: project, noteable: issue,
          note: "Sample comment with Git commit: #{git_sha}")
      end

      it 'migrate comments' do
        subject.migrate_notes
        my_comment.reload
        expect(my_comment.note).to eq("Sample comment with Git commit: #{hg_sha}")
      end
    end
  end

  context 'Merge Requests' do
    let!(:merge_request) do
      create(:merge_request,
        source_project: project,
        source_branch: 'topic/default/feature',
        target_project: project,
        target_branch: 'branch/default',
        description: "Sample description with Git commit: #{git_sha}",
        title: "Sample title for Git commit: #{git_sha}")
    end

    it 'migrate title, description, and merge_request_diff_commits' do
      expect(merge_request.all_commits.first.sha).to eq(git_sha)

      # migrate merge request
      subject.migrate_merge_requests
      merge_request.reload

      expect(merge_request.all_commits.first.sha).to eq(hg_sha)
      expect(merge_request.description).to eq("Sample description with Git commit: #{hg_sha}")
      expect(merge_request.title).to eq("Sample title for Git commit: #{hg_sha}")
    end

    context 'Notes' do
      let!(:my_comment) do
        create(:note_on_merge_request, project: project, noteable: merge_request,
          note: "Sample comment with Git commit: #{git_sha}")
      end

      it 'migrate comments' do
        subject.migrate_notes
        my_comment.reload
        expect(my_comment.note).to eq("Sample comment with Git commit: #{hg_sha}")
      end
    end

    context 'DiffNote' do
      let(:path) { "file2"}
      let!(:position) do
        Gitlab::Diff::Position.new(
          old_path: path,
          new_path: path,
          old_line: nil,
          new_line: 1,
          diff_refs: merge_request.diff_refs
        )
      end

      shared_examples 'DiffNote migration' do
        let!(:note) do
          create(:diff_note_on_merge_request,
                 project: project,
                 position: position,
                 noteable: merge_request,
                 note: "This reminds me of #{git_base_sha}")
        end

        it "migrates the note" do
          line_code_before = note.line_code
          subject.migrate_notes
          note.reload
          expect(note.note).to eq("This reminds me of #{hg_base_sha}")
          expect(note.line_code).to eq(line_code_before)
        end
      end

      context "when rugged Git implementation is enabled", :enable_rugged do
        it_behaves_like 'DiffNote migration'
      end

      context "when rugged Git implementation is not enabled" do
        it_behaves_like 'DiffNote migration'
      end
    end
  end

  context 'Pipelines' do
    let!(:pipeline) do
      create(:ci_pipeline,
        project: project,
        sha: git_sha,
        ref: 'topic/default/feature',
        status: 'success')
    end

    it 'migrate pipelines' do
      subject.migrate_pipelines
      pipeline.reload
      expect(pipeline.sha).to eq(hg_sha)
    end
  end

  context 'Releases' do
    let!(:release) do
      create(:release,
        project: project,
        description: "Sample description with Git commit: #{git_sha}")
    end

    it 'migrate description' do
      subject.migrate_releases
      release.reload
      expect(release.description).to eq("Sample description with Git commit: #{hg_sha}")
    end
  end

  describe '#vcs_type_enforcer' do
    let!(:project) { create(:native_hg_project, :hg_repository) }

    subject { described_class.new(project) }

    it 'force repository to be HgGitRepository' do
      expect(project.vcs_type).to eq("hg")
      subject.vcs_type_enforcer do
        expect(project.vcs_type).to eq("hg_git")
      end
      expect(project.vcs_type).to eq("hg")
    end
  end
end
